"""
Module implements a wrapper around the Power BI Rest API Group operations.

Full Group Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/groups

"""

from pbipy.resources import Resource
from pbipy import _utils

from requests import Session


class Group(Resource):
    """
    A Power BI Group (better know as a Workspace).

    Users should initialize a `Group` object by calling the `group()` method
    on the `PowerBI` client, rather than creating directly.

    Examples
    --------
    Retrieving a `Group` object using a `pbi` client object.

    ```
    >>> my_group = pbi.group("a2f89923-421a-464e-bf4c-25eab39bb09f")
    ```

    Retrieve a list of users with access to a Group.

    ```
    >>> my_group = pbi.group("a2f89923-421a-464e-bf4c-25eab39bb09f")
    >>> group_users = my_group.users()
    ```

    """

    _REPR = [
        "id",
        "name",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        raw: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(id, session, **kwargs)

        self.base_path = f"{self.BASE_URL}/groups/{self.id}"

        if raw:
            self._load_from_raw(raw)

    def add_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
        display_name: str = None,
        email_address: str = None,
        graph_id: str = None,
        profile: dict = None,
        user_type: str = None,
    ) -> None:
        """
        Grants the specified user the specified permissions to the workspace.

        Parameters
        ----------
        `identifier` : `str`
            Identifier of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The access right (permission level) that a user has on the
            workspace, e.g., "Admin", "Contributor", "Member", "None",
            or "Viewer".
        `display_name` : `str`, optional
            Display name of the principal.
        `email_address` : `str`, optional
            Email address of the user.
        `graph_id` : `str`, optional
            Identifier of the principal in Microsoft Graph. Only available
            for admin APIs.
        `profile` : `dict`, optional
            A Power BI service principal profile. Only relevant for Power
            BI Embedded multi-tenancy solution.
        `user_type` : `str`, optional
            Type of the user.

        """

        payload = {
            "identifier": identifier,
            "groupUserAccessRight": access_right,
            "principalType": principal_type,
            "displayName": display_name,
            "emailAddress": email_address,
            "graphId": graph_id,
            "profile": profile,
            "userType": user_type,
        }

        prepared_payload = _utils.remove_no_values(payload)
        resource = self.base_path + "/users"

        _utils.post(resource, self.session, prepared_payload)

    def delete_user(
        self,
        user: str,
        profile: str = None,
    ) -> None:
        """
        Deletes the specified user permissions from the specified workspace.

        Parameters
        ----------
        `user` : `str`
            The email address of the user or object ID of the service principal
            to delete.
        `profile` : `str`, optional
            The service principal profile ID to delete.

        """

        if profile:
            resource = self.base_path + f"/users/{user}?profileId={profile}"
        else:
            resource = self.base_path + f"/users/{user}"

        _utils.delete(resource, self.session)

    def update_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
        display_name: str = None,
        email_address: str = None,
        graph_id: str = None,
        profile: dict = None,
        user_type: str = None,
    ) -> None:
        """
        Updates the specified user permissions on the workspace.

        Parameters
        ----------
        `identifier` : `str`
            Identifier of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The access right (permission level) that a user has on the
            workspace, e.g., "Admin", "Contributor", "Member", "None",
            or "Viewer".
        `display_name` : `str`, optional
            Display name of the principal.
        `email_address` : `str`, optional
            Email address of the user.
        `graph_id` : `str`, optional
            Identifier of the principal in Microsoft Graph. Only available
            for admin APIs.
        `profile` : `dict`, optional
            A Power BI service principal profile. Only relevant for Power
            BI Embedded multi-tenancy solution.
        `user_type` : `str`, optional
            Type of the user.

        """

        payload = {
            "identifier": identifier,
            "groupUserAccessRight": access_right,
            "principalType": principal_type,
            "displayName": display_name,
            "emailAddress": email_address,
            "graphId": graph_id,
            "profile": profile,
            "userType": user_type,
        }

        prepared_payload = _utils.remove_no_values(payload)
        resource = self.base_path + "/users"

        _utils.put(resource, self.session, prepared_payload)

    def update(
        self,
        name: str = None,
        default_dataset_storage_format: str = None,
    ) -> "Group":
        """
        Update properties for this Group on the Power BI instance. First attempts
        to update the Group on the Power BI instance, and then reflects these
        changes in the Group.

        Parameters
        ----------
        `name` : `str`, optional
            The new name for the group.
        `default_dataset_storage_format` : `str`, optional
            The new Default Dataset Storage Format for the group.

        Returns
        -------
        `Group`
            The group with the specified values updated.

        Raises
        ------
        `ValueError`
            If no values to update were provided.

        """

        request_body = {
            "name": name,
            "defaultDatasetStorageFormat": default_dataset_storage_format,
        }
        request_body = _utils.remove_no_values(request_body)

        if not request_body:
            raise ValueError(
                "No options were provided to update. Please specify an option to update."
            )

        _utils.patch(
            self.base_path,
            self.session,
            request_body,
        )

        # Update existing group to reflect changes
        self.raw.update(request_body)

        for k, v in request_body.items():
            attr = _utils.to_identifier(k)
            attr = _utils.to_snake_case(attr)
            setattr(self, attr, v)

        return self

    def users(
        self,
        skip: int = None,
        top: int = None,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the workspace.

        Parameters
        ----------
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns onl the first n results.
        Returns
        -------
        `list[dict]`
            List of users that have access to the workspace.

        """

        params = {
            "$skip": skip,
            "$top": top,
        }

        prepared_params = _utils.remove_no_values(params)
        resource = self.base_path + "/users"

        raw = _utils.get_raw(
            resource,
            self.session,
            params=prepared_params,
        )

        return raw



"""Embed token definition."""

from types import SimpleNamespace


class EmbedToken(SimpleNamespace):
    """
    A Power BI embed token.

    """

    def __init__(
        self,
        token: str,
        token_id: str,
        expiration: str,
    ):
        super().__init__(
            token=token,
            token_id=token_id,
            expiration=expiration,
        )



import pytest
import requests
import responses
from requests.exceptions import HTTPError
from responses import matchers

from pbipy.datasets import Dataset, DatasetRefreshError


@responses.activate
def test_get_refresh_history():
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes?$top=2",
        body="""
        {
            "value": [
                {
                "refreshType": "ViaApi",
                "startTime": "2017-06-13T09:25:43.153Z",
                "endTime": "2017-06-13T09:31:43.153Z",
                "status": "Completed",
                "requestId": "9399bb89-25d1-44f8-8576-136d7e9014b1"
                }
            ]
        }
        """,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    refresh_history = dataset.refresh_history(top=2)

    assert isinstance(refresh_history, list)
    assert all(isinstance(refresh, dict) for refresh in refresh_history)


@responses.activate
def test_get_refresh_history_calls_correct_url():
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes?$top=2",
        body='{"value": []}',
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.refresh_history(top=2)


@responses.activate
def test_get_refresh_history_raises_http_error():
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes",
        body='{"value": []}',
        content_type="application/json",
        status=400,
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    with pytest.raises(HTTPError):
        dataset.refresh_history()


@responses.activate
def test_post_dataset_user_call():
    json_params = {
        "datasetUserAccessRight": "Read",
        "identifier": "john@contoso.com",
        "principalType": "User",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.add_user("john@contoso.com", "User", "Read")


@responses.activate
def test_post_dataset_raises_http_error():
    json_params = {
        "datasetUserAccessRight": "Read",
        "identifier": "john@contoso.com",
        "principalType": "User",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        status=400,
        json={
            "error": {
                "code": "InvalidRequest",
                "message": "Parameter identifier is missing or invalid",
            }
        },
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    with pytest.raises(HTTPError):
        dataset.add_user("john@contoso.com", "User", "Read")


@responses.activate
def test_bind_to_gateway_with_data_source_object_ids():
    json_params = {
        "gatewayObjectId": "1f69e798-5852-4fdd-ab01-33bb14b6e934",
        "datasourceObjectIds": [
            "dc2f2dac-e5e2-4c37-af76-2a0bc10f16cb",
            "3bfe5d33-ab7d-4d24-b0b5-e2bb8eb01cf5",
        ],
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.BindToGateway",
        match=[matchers.json_params_matcher(json_params)],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    datasource_object_ids = [
        "dc2f2dac-e5e2-4c37-af76-2a0bc10f16cb",
        "3bfe5d33-ab7d-4d24-b0b5-e2bb8eb01cf5",
    ]

    dataset.bind_to_gateway(
        gateway_object_id="1f69e798-5852-4fdd-ab01-33bb14b6e934",
        datasource_object_ids=datasource_object_ids,
    )


@responses.activate
def test_bind_to_gateway_without_data_source_object_ids():
    json_params = {
        "gatewayObjectId": "1f69e798-5852-4fdd-ab01-33bb14b6e934",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.BindToGateway",
        match=[matchers.json_params_matcher(json_params)],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.bind_to_gateway(gateway_object_id="1f69e798-5852-4fdd-ab01-33bb14b6e934")


@responses.activate
def test_cancel_refresh():
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/fdb91b8f-0a9b-44c1-b6c0-0cb185c6ebfb/datasets/f7fc6510-e151-42a3-850b-d0805a391db0/refreshes/87f31ef7-1e3a-4006-9b0b-191693e79e9e"
    )

    dataset = Dataset(
        id="f7fc6510-e151-42a3-850b-d0805a391db0",
        session=requests.Session(),
        group_id="fdb91b8f-0a9b-44c1-b6c0-0cb185c6ebfb",
    )

    dataset.cancel_refresh("87f31ef7-1e3a-4006-9b0b-191693e79e9e")


@responses.activate
def test_discover_gateways():
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.DiscoverGateways",
        body="""
        {
            "value": [
                {
                "id": "1f69e798-5852-4fdd-ab01-33bb14b6e934",
                "name": "ContosoGateway",
                "type": "Resource",
                "publicKey": {
                    "exponent": "AQAB",
                    "modulus": "o6j2....cLk="
                }
                }
            ]
        }
        """,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    gateways = dataset.discover_gateways()

    assert isinstance(gateways, list)
    assert all(isinstance(gateway, dict) for gateway in gateways)


@responses.activate
def test_discover_gateways_call():
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.DiscoverGateways",
        body='{"value": []}',
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )
    dataset.discover_gateways()


@responses.activate
def test_execute_queries_call(execute_queries):
    json_params = {
        "queries": [
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
        ],
        "impersonatedUserName": "someuser@mycompany.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/executeQueries",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=execute_queries,
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    query = "EVALUATE VALUES(MyTable)"

    dataset.execute_queries(
        query,
        impersonated_user_name="someuser@mycompany.com",
    )


@responses.activate
def test_execute_queries_call_multi_query(execute_queries):
    json_params = {
        "queries": [
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
        ],
        "impersonatedUserName": "someuser@mycompany.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/executeQueries",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=execute_queries,
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    queries = [
        "EVALUATE VALUES(MyTable)",
        "EVALUATE VALUES(MyTable)",
    ]

    dataset.execute_queries(
        queries,
        impersonated_user_name="someuser@mycompany.com",
    )


@responses.activate
def test_execute_queries_call_result(execute_queries):
    json_params = {
        "queries": [
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
        ],
        "impersonatedUserName": "someuser@mycompany.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/executeQueries",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=execute_queries,
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    queries = "EVALUATE VALUES(MyTable)"

    result = dataset.execute_queries(
        queries,
        impersonated_user_name="someuser@mycompany.com",
    )

    assert isinstance(result, dict)
    assert "results" in result


@responses.activate
def test_users_call(get_dataset_users):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        body=get_dataset_users,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.users()


@responses.activate
def test_users_result(get_dataset_users):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        body=get_dataset_users,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    users = dataset.users()

    assert isinstance(users, list)
    assert len(users) == 3


@responses.activate
def test_datasources_call(get_datasources):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources",
        body=get_datasources,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.datasources()


@responses.activate
def test_datasources_result(get_datasources):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources",
        body=get_datasources,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    datasources = dataset.datasources()

    assert isinstance(datasources, list)
    assert len(datasources) == 3


@responses.activate
def test_refresh_schedule_call(get_refresh_schedule):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshSchedule",
        body=get_refresh_schedule,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.refresh_schedule()


@responses.activate
def test_refresh_schedule_result(get_refresh_schedule):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshSchedule",
        body=get_refresh_schedule,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    refresh_schedule = dataset.refresh_schedule()

    assert isinstance(refresh_schedule, dict)


@responses.activate
def test_refresh_schedule_direct_query_call(get_direct_query_refresh_schedule):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/directQueryRefreshSchedule",
        body=get_direct_query_refresh_schedule,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.refresh_schedule(direct_query=True)


@responses.activate
def test_refresh_schedule_direct_query_result(get_direct_query_refresh_schedule):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/directQueryRefreshSchedule",
        body=get_direct_query_refresh_schedule,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    refresh_schedule = dataset.refresh_schedule(direct_query=True)

    assert isinstance(refresh_schedule, dict)


@responses.activate
def test_parameters_call(get_parameters):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/parameters",
        body=get_parameters,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.parameters()


@responses.activate
def test_parameters_result(get_parameters):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/parameters",
        body=get_parameters,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    parameters = dataset.parameters()

    assert isinstance(parameters, list)
    assert all(isinstance(parameter, dict) for parameter in parameters)
    assert len(parameters) == 6


@responses.activate
def test_refresh_details_call(get_refresh_execution_details):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/f7fc6510-e151-42a3-850b-d0805a391db0/refreshes/87f31ef7-1e3a-4006-9b0b-191693e79e9e",
        body=get_refresh_execution_details,
        content_type="application/json",
    )

    dataset = Dataset(
        id="f7fc6510-e151-42a3-850b-d0805a391db0",
        session=requests.Session(),
    )

    dataset.refresh_details("87f31ef7-1e3a-4006-9b0b-191693e79e9e")


@responses.activate
def test_refresh_details_result(get_refresh_execution_details):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/f7fc6510-e151-42a3-850b-d0805a391db0/refreshes/87f31ef7-1e3a-4006-9b0b-191693e79e9e",
        body=get_refresh_execution_details,
        content_type="application/json",
    )

    dataset = Dataset(
        id="f7fc6510-e151-42a3-850b-d0805a391db0",
        session=requests.Session(),
    )

    refresh_detail = dataset.refresh_details("87f31ef7-1e3a-4006-9b0b-191693e79e9e")

    assert isinstance(refresh_detail, dict)
    assert len(refresh_detail["objects"]) == 18


@pytest.fixture
def sleepless(monkeypatch):
    """Patches time.sleep to make unit tests complete without delay."""

    import time

    def sleep(seconds):
        pass

    monkeypatch.setattr(time, "sleep", sleep)


@responses.activate
def test_refresh_and_wait_success(
    get_refresh_execution_details_in_progress,
    get_refresh_execution_details,
    sleepless,
):
    dataset_id = "cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    request_id = "03f22bb5-2e98-4ae8-8113-329bec3987b1"

    dataset = Dataset(
        id=dataset_id,
        session=requests.Session(),
    )

    json_params = {"type": "Full"}

    responses.post(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        headers={"RequestId": request_id},
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details,
        content_type="application/json",
    )

    dataset.refresh_and_wait(type="Full", check_interval=60)
    assert len(responses.calls) == 3


@responses.activate
def test_refresh_and_wait_success_longer(
    get_refresh_execution_details_in_progress,
    get_refresh_execution_details,
    sleepless,
):
    dataset_id = "cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    request_id = "03f22bb5-2e98-4ae8-8113-329bec3987b1"

    dataset = Dataset(
        id=dataset_id,
        session=requests.Session(),
    )

    json_params = {"type": "Full", "retryCount": 3}

    responses.post(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        headers={"RequestId": request_id},
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details,
        content_type="application/json",
    )

    dataset.refresh_and_wait(type="Full", retry_count=3)
    assert len(responses.calls) == 5


def test_refresh_and_wait_raises_ValueError():
    dataset_id = "cfafbeb1-8037-4d0c-896e-a46fb27ff229"

    dataset = Dataset(
        id=dataset_id,
        session=requests.Session(),
    )

    with pytest.raises(ValueError):
        dataset.refresh_and_wait()


@responses.activate
def test_refresh_and_wait_failed_raises_RefreshDatasetError(
    get_refresh_execution_details_in_progress,
    get_refresh_execution_details_failed,
    sleepless,
):
    dataset_id = "cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    request_id = "03f22bb5-2e98-4ae8-8113-329bec3987b1"

    dataset = Dataset(
        id=dataset_id,
        session=requests.Session(),
    )

    json_params = {"type": "Full"}

    responses.post(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        headers={"RequestId": request_id},
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_failed,
        content_type="application/json",
    )

    with pytest.raises(DatasetRefreshError):
        dataset.refresh_and_wait(type="Full", check_interval=60)

    assert len(responses.calls) == 3


@responses.activate
def test_refresh_and_wait_cancelled_raises_RefreshDatasetError(
    get_refresh_execution_details_in_progress,
    get_refresh_execution_details_cancelled,
    sleepless,
):
    dataset_id = "cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    request_id = "03f22bb5-2e98-4ae8-8113-329bec3987b1"

    dataset = Dataset(
        id=dataset_id,
        session=requests.Session(),
    )

    json_params = {"type": "Full"}

    responses.post(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        headers={"RequestId": request_id},
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_in_progress,
        content_type="application/json",
    )

    responses.get(
        f"https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/refreshes/{request_id}",
        body=get_refresh_execution_details_cancelled,
        content_type="application/json",
    )

    with pytest.raises(DatasetRefreshError):
        dataset.refresh_and_wait(type="Full", check_interval=60)

    assert len(responses.calls) == 3


@responses.activate
def test_update_user_call():
    json_params = {
        "datasetUserAccessRight": "Read",
        "identifier": "john@contoso.com",
        "principalType": "User",
    }

    responses.put(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.update_user("john@contoso.com", "User", "Read")


@responses.activate
def test_update_user_raises_http_error():
    json_params = {
        "datasetUserAccessRight": "Read",
        "identifier": "john@contoso.com",
        "principalType": "User",
    }

    responses.put(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        status=400,
        json={
            "error": {
                "code": "InvalidRequest",
                "message": "Parameter identifier is missing or invalid",
            }
        },
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    with pytest.raises(HTTPError):
        dataset.update_user("john@contoso.com", "User", "Read")


@responses.activate
def test_refresh_call_simple():
    request_id = "request_id"

    json_parms = {
        "notifyOption": "MailOnFailure",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes",
        match=[
            matchers.json_params_matcher(json_parms),
        ],
        headers={"requestId": request_id},
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    refresh_id = dataset.refresh(notify_option="MailOnFailure")

    assert refresh_id == request_id


@responses.activate
def test_refresh_call_complex():
    request_id = "request_id"

    json_parms = {
        "notifyOption": "MailOnFailure",
        "retryCount": 3,
        "type": "full",
        "commitMode": "transactional",
        "objects": [
            {
                "table": "Customer",
                "partition": "Robert",
            },
        ],
        "applyRefreshPolicy": False,
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes",
        match=[
            matchers.json_params_matcher(json_parms),
        ],
        headers={"requestId": request_id},
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    refresh_id = dataset.refresh(
        notify_option="MailOnFailure",
        retry_count=3,
        type="full",
        commit_mode="transactional",
        apply_refresh_policy=False,
        objects=[
            {
                "table": "Customer",
                "partition": "Robert",
            },
        ],
    )

    assert refresh_id == request_id


@responses.activate
def test_take_over_call():
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.TakeOver",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.take_over()


@responses.activate
def test_take_over_raises_type_error():
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.TakeOver",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    with pytest.raises(TypeError):
        dataset.take_over()


@responses.activate
def test_update_call():
    json_params = {"targetStorageMode": "PremiumFiles"}

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    dataset.update(target_storage_mode="PremiumFiles")


@responses.activate
def test_update_datasources_single():
    json_params = {
        "updateDetails": [
            {
                "datasourceSelector": {
                    "datasourceType": "Sql",
                    "connectionDetails": {
                        "server": "My-Sql-Server",
                        "database": "My-Sql-Database",
                    },
                },
                "connectionDetails": {
                    "server": "New-Sql-Server",
                    "database": "New-Sql-Database",
                },
            },
        ]
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.UpdateDatasources",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    update_details = {
        "datasourceSelector": {
            "datasourceType": "Sql",
            "connectionDetails": {
                "server": "My-Sql-Server",
                "database": "My-Sql-Database",
            },
        },
        "connectionDetails": {
            "server": "New-Sql-Server",
            "database": "New-Sql-Database",
        },
    }

    dataset.update_datasources(update_details=update_details)


@responses.activate
def test_update_datasources_multi():
    json_params = {
        "updateDetails": [
            {
                "datasourceSelector": {
                    "datasourceType": "Sql",
                    "connectionDetails": {
                        "server": "My-Sql-Server",
                        "database": "My-Sql-Database",
                    },
                },
                "connectionDetails": {
                    "server": "New-Sql-Server",
                    "database": "New-Sql-Database",
                },
            },
            {
                "datasourceSelector": {
                    "datasourceType": "OData",
                    "connectionDetails": {
                        "url": "http://services.odata.org/V4/Northwind/Northwind.svc"
                    },
                },
                "connectionDetails": {
                    "url": "http://services.odata.org/V4/Odata/Northwind.svc"
                },
            },
        ]
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.UpdateDatasources",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    update_details = [
        {
            "datasourceSelector": {
                "datasourceType": "Sql",
                "connectionDetails": {
                    "server": "My-Sql-Server",
                    "database": "My-Sql-Database",
                },
            },
            "connectionDetails": {
                "server": "New-Sql-Server",
                "database": "New-Sql-Database",
            },
        },
        {
            "datasourceSelector": {
                "datasourceType": "OData",
                "connectionDetails": {
                    "url": "http://services.odata.org/V4/Northwind/Northwind.svc"
                },
            },
            "connectionDetails": {
                "url": "http://services.odata.org/V4/Odata/Northwind.svc"
            },
        },
    ]

    dataset.update_datasources(update_details=update_details)


@responses.activate
def test_update_refresh_schedule():
    json_params = {
        "value": {
            "days": [
                "Sunday",
                "Tuesday",
                "Friday",
                "Saturday",
            ],
            "times": [
                "07:00",
                "11:30",
                "16:00",
                "23:30",
            ],
            "localTimeZoneId": "UTC",
        }
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshSchedule",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.update_refresh_schedule(
        days=[
            "Sunday",
            "Tuesday",
            "Friday",
            "Saturday",
        ],
        times=[
            "07:00",
            "11:30",
            "16:00",
            "23:30",
        ],
        local_time_zone_id="UTC",
    )


@responses.activate
def test_update_refresh_schedule_direct_query():
    json_params = {
        "value": {
            "days": [
                "Sunday",
                "Tuesday",
                "Friday",
                "Saturday",
            ],
            "frequency": 15,
            "localTimeZoneId": "UTC",
        }
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/directQueryRefreshSchedule",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    dataset.update_refresh_schedule(
        direct_query=True,
        days=[
            "Sunday",
            "Tuesday",
            "Friday",
            "Saturday",
        ],
        frequency=15,
        local_time_zone_id="UTC",
    )


@responses.activate
def test_update_refresh_schedule_raises():
    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/directQueryRefreshSchedule",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )
    with pytest.raises(ValueError):
        dataset.update_refresh_schedule()


@responses.activate
def test_update_parameters_single():
    json_params = {
        "updateDetails": [
            {
                "name": "DatabaseName",
                "newValue": "NewDB",
            },
        ]
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.UpdateParameters",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    update_details = {
        "name": "DatabaseName",
        "newValue": "NewDB",
    }

    dataset.update_parameters(update_details)


@responses.activate
def test_update_parameters_multiple():
    json_params = {
        "updateDetails": [
            {
                "name": "DatabaseName",
                "newValue": "NewDB",
            },
            {
                "name": "MaxId",
                "newValue": "5678",
            },
        ]
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.UpdateParameters",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    update_details = [
        {
            "name": "DatabaseName",
            "newValue": "NewDB",
        },
        {
            "name": "MaxId",
            "newValue": "5678",
        },
    ]

    dataset.update_parameters(update_details)


@responses.activate
def test_refresh():
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/refreshes",
        headers={"RequestId": "03f22bb5-2e98-4ae8-8113-329bec3987b1"},
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    refresh_id = dataset.refresh()

    assert refresh_id == "03f22bb5-2e98-4ae8-8113-329bec3987b1"



from datetime import datetime

import pytest
import requests
import responses
from responses import matchers
from responses.registries import OrderedRegistry

from pbipy.apps import App
from pbipy.dashboards import Dashboard, Tile
from pbipy.dataflows import Dataflow
from pbipy.datasets import Dataset
from pbipy.groups import Group
from pbipy.reports import Report


@pytest.fixture
def admin(powerbi):
    return powerbi.admin()


@responses.activate
def test_app_users(admin, get_app_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        body=get_app_users_as_admin,
        content_type="application/json",
    )

    users = admin.app_users("f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_apps(admin, get_apps_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/apps",
        body=get_apps_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    apps = admin.apps()


@responses.activate
def test_apps_params(admin, get_apps_as_admin):
    params = {"$top": 1}
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/apps",
        body=get_apps_as_admin,
        match=[
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    apps = admin.apps(top=1)


@responses.activate
def test_apps_result(admin, get_apps_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/apps",
        body=get_apps_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    apps = admin.apps()

    assert isinstance(apps, list)
    assert all(isinstance(app, App) for app in apps)


@responses.activate
def test_dashboards(admin, get_dashboards_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards",
        body=get_dashboards_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    dashboards = admin.dashboards()


@responses.activate
def test_dashboards_with_params(admin, get_dashboards_as_admin):
    params = {
        "$skip": 1,
        "$top": 3,
        "$filter": "contains(name,'marketing')",
    }

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards",
        body=get_dashboards_as_admin,
        match=[
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    dashboards = admin.dashboards(
        skip=1,
        top=3,
        filter="contains(name,'marketing')",
    )


@responses.activate
def test_dashboards_result(admin, get_dashboards_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards",
        body=get_dashboards_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    dashboards = admin.dashboards()

    assert isinstance(dashboards, list)
    assert all(isinstance(dashboard, Dashboard) for dashboard in dashboards)


@responses.activate
def test_dashboards_with_group(admin, get_dashboards_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards",
        body=get_dashboards_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )
    group = Group(
        id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )
    dashboards = admin.dashboards(group=group)


# TODO: Need to add response body for ExportDataflowAsAdmin endpoint
@responses.activate
def test_dataflow(admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/export",
        body="{}",
        content_type="application/json",
    )

    admin.dataflow(dataflow="928228ba-008d-4fd9-864a-92d2752ee5ce")


@responses.activate
def test_dataflow_datasources(admin, get_dataflow_datasources_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dataflows/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources",
        body=get_dataflow_datasources_as_admin,
        content_type="application/json",
    )

    datasources = admin.dataflow_datasources(
        dataflow="cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    )

    assert isinstance(datasources, list)
    assert all(isinstance(datasource, dict) for datasource in datasources)


@responses.activate
def test_dataflow_users(admin, get_dataflow_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dataflows/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        body=get_dataflow_users_as_admin,
        content_type="application/json",
    )

    users = admin.dataflow_users(dataflow="cfafbeb1-8037-4d0c-896e-a46fb27ff229")

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_dataflow_upstream_dataflows(admin, get_upstream_dataflows_in_group_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dataflows/cfafbeb1-8037-4d0c-896e-a46fb27ff229/upstreamDataflows",
        body=get_upstream_dataflows_in_group_as_admin,
        content_type="application/json",
    )

    upstream_dataflows = admin.dataflow_upstream_dataflows(
        dataflow="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(upstream_dataflows, list)
    assert all(
        isinstance(upstream_dataflow, dict) for upstream_dataflow in upstream_dataflows
    )


@responses.activate
def test_dataset_upstream_dataflows(
    admin,
    get_dataset_to_dataflow_links_in_group_as_admin,
):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/upstreamDataflows",
        body=get_dataset_to_dataflow_links_in_group_as_admin,
        content_type="application/json",
    )

    upstream_dataflows = admin.datasets_upstream_dataflows(
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )

    assert isinstance(upstream_dataflows, list)
    assert all(
        isinstance(upstream_dataflow, dict) for upstream_dataflow in upstream_dataflows
    )


@responses.activate
def test_dataset_users(admin, get_dataset_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        body=get_dataset_users_as_admin,
        content_type="application/json",
    )

    dataset = Dataset(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    users = admin.dataset_users(dataset=dataset)

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_datasets(admin, get_datasets_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/datasets",
        body=get_datasets_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    datasets = admin.datasets()


@responses.activate
def test_datasets_with_params(admin, get_datasets_as_admin):
    params = {
        "$top": 2,
        "$filter": "contains(name,'marketing')",
    }

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/datasets",
        body=get_datasets_as_admin,
        match=[
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    datasets = admin.datasets(filter="contains(name,'marketing')", top=2)


@responses.activate
def test_datasets_results(admin, get_datasets_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/datasets",
        body=get_datasets_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    datasets = admin.datasets()

    assert isinstance(datasets, list)
    assert all(isinstance(dataset, Dataset) for dataset in datasets)
    assert datasets[0].group_id is not None
    assert datasets[0].group_id == datasets[0].workspace_id


@responses.activate
def test_datasets_with_group(admin, get_datasets_in_group_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets",
        body=get_datasets_in_group_as_admin,
        content_type="application/json",
    )

    datasets = admin.datasets(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert datasets[0].group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert not hasattr(datasets[0], "workspace_id")


@responses.activate
def test_datasets_with_group_obj(admin, get_datasets_in_group_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets",
        body=get_datasets_in_group_as_admin,
        content_type="application/json",
    )
    group = Group(
        id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )
    datasets = admin.datasets(group=group)

    assert datasets[0].group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert not hasattr(datasets[0], "workspace_id")


@responses.activate
def test_dataflows(admin, get_dataflows_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dataflows",
        body=get_dataflows_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    dataflows = admin.dataflows()

    assert isinstance(dataflows, list)
    assert all(isinstance(dataflow, Dataflow) for dataflow in dataflows)

    assert dataflows[0].id == "bd32e5c0-363f-430b-a03b-5535a4804b9b"
    assert dataflows[0].name
    assert dataflows[0].description
    assert dataflows[0].model_url
    assert dataflows[0].configured_by
    assert dataflows[0].group_id == dataflows[0].workspace_id


@responses.activate
def test_dataflows_with_group(admin, get_dataflows_in_group_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dataflows",
        body=get_dataflows_in_group_as_admin,
        match=[
            matchers.query_param_matcher(None),
        ],
        content_type="application/json",
    )

    dataflows = admin.dataflows(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(dataflows, list)
    assert all(isinstance(dataflow, Dataflow) for dataflow in dataflows)

    assert dataflows[0].id
    assert dataflows[0].group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert not hasattr(dataflows[0], "workspace_id")


@responses.activate
def test_dataflows_with_params(admin, get_dataflows_as_admin):
    params = {"$top": 1, "$skip": 3}

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dataflows",
        body=get_dataflows_as_admin,
        match=[
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    dataflows = admin.dataflows(top=1, skip=3)


@responses.activate
def test_add_encryption_key(admin, add_power_bi_encryption_key):
    json_params = {
        "name": "Contoso Sales",
        "keyVaultKeyIdentifier": "https://contoso-vault2.vault.azure.net/keys/ContosoKeyVault/b2ab4ba1c7b341eea5ecaaa2wb54c4d2",
        "activate": True,
        "isDefault": True,
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/tenantKeys",
        body=add_power_bi_encryption_key,
        match=[
            matchers.json_params_matcher(json_params),
        ],
        content_type="application/json",
    )

    encryption_key = admin.add_encryption_key(
        name="Contoso Sales",
        key_vault_identifier="https://contoso-vault2.vault.azure.net/keys/ContosoKeyVault/b2ab4ba1c7b341eea5ecaaa2wb54c4d2",
        activate=True,
        is_default=True,
    )

    assert isinstance(encryption_key, dict)


@responses.activate
def test_dashboard_subscriptions(admin, get_dashboard_subscriptions_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards/69ffaa6c-b36d-4d01-96f5-1ed67c64d4af/subscriptions",
        body=get_dashboard_subscriptions_as_admin,
        content_type="application/json",
    )

    subscriptions = admin.dashboard_subscriptions(
        dashboard="69ffaa6c-b36d-4d01-96f5-1ed67c64d4af"
    )

    assert isinstance(subscriptions, list)
    assert all(isinstance(subscription, dict) for subscription in subscriptions)


@responses.activate
def test_dashboard_users(admin, get_dashboard_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards/69ffaa6c-b36d-4d01-96f5-1ed67c64d4af/users",
        body=get_dashboard_users_as_admin,
        content_type="application/json",
    )

    users = admin.dashboard_users(dashboard="69ffaa6c-b36d-4d01-96f5-1ed67c64d4af")

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_dashboard_tiles(admin, get_tiles_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/dashboards/69ffaa6c-b36d-4d01-96f5-1ed67c64d4af/tiles",
        body=get_tiles_as_admin,
        content_type="application/json",
    )

    dashboard = Dashboard(
        id="69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        session=requests.Session(),
    )

    tiles = admin.dashboard_tiles(dashboard)

    assert isinstance(tiles, list)
    assert all(isinstance(tile, Tile) for tile in tiles)


@responses.activate
def test_dataset_datasources(admin, get_datasources_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources",
        body=get_datasources_as_admin,
        content_type="application/json",
    )

    datasources = admin.dataset_datasources("cfafbeb1-8037-4d0c-896e-a46fb27ff229")

    assert isinstance(datasources, list)
    assert all(isinstance(datasource, dict) for datasource in datasources)


@responses.activate
def test_encryption_keys(admin, get_power_bi_encryption_keys):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/tenantKeys",
        body=get_power_bi_encryption_keys,
        content_type="application/json",
    )

    encryption_keys = admin.encryption_keys()

    assert isinstance(encryption_keys, list)
    assert all(isinstance(encryption_key, dict) for encryption_key in encryption_keys)


@responses.activate
def test_add_group_user(admin):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "groupUserAccessRight": "Admin",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/206d27ca-94e8-4a69-855b-5c32bdd458a8/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    admin.add_group_user(
        group="206d27ca-94e8-4a69-855b-5c32bdd458a8",
        identifier="john@contoso.com",
        principal_type="User",
        group_user_access_right="Admin",
    )


@responses.activate
def test_add_group_user_with_optional_params(admin):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "groupUserAccessRight": "Admin",
        "emailAddress": "john@contoso.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/206d27ca-94e8-4a69-855b-5c32bdd458a8/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    admin.add_group_user(
        group="206d27ca-94e8-4a69-855b-5c32bdd458a8",
        identifier="john@contoso.com",
        principal_type="User",
        group_user_access_right="Admin",
        email_address="john@contoso.com",
    )


@responses.activate
def test_delete_group_user(admin):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f7d76f5a-7190-43c6-bf12-7a135c6c2d69/users/john@contoso.com"
    )

    admin.delete_group_user(
        "f7d76f5a-7190-43c6-bf12-7a135c6c2d69",
        "john@contoso.com",
    )


@responses.activate
def test_delete_group_user_with_params(admin):
    params = {"profileId": "154aef10-47b8-48c4-ab97-f0bf9d5f8fcf"}

    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f7d76f5a-7190-43c6-bf12-7a135c6c2d69/users/john@contoso.com",
        match=[
            matchers.query_param_matcher(params),
        ],
    )

    admin.delete_group_user(
        "f7d76f5a-7190-43c6-bf12-7a135c6c2d69",
        "john@contoso.com",
        profile_id="154aef10-47b8-48c4-ab97-f0bf9d5f8fcf",
    )


@responses.activate
def test_group(admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/a2f89923-421a-464e-bf4c-25eab39bb09f",
        body="""{"id": "a2f89923-421a-464e-bf4c-25eab39bb09f"}""",
        content_type="application/json",
    )

    group = admin.group("a2f89923-421a-464e-bf4c-25eab39bb09f")

    assert isinstance(group, Group)


@responses.activate
def test_groups(admin, get_groups_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups?$top=1",
        body=get_groups_as_admin,
        content_type="application/json",
    )

    groups = admin.groups(top=1)

    assert isinstance(groups, list)
    assert all(isinstance(group, Group) for group in groups)


@responses.activate
def test_groups_without_top(admin, get_groups_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups",
        body=get_groups_as_admin,
        content_type="application/json",
        match=[
            matchers.query_param_matcher({"$top": 5000}),
        ],
    )

    groups = admin.groups()

    assert isinstance(groups, list)
    assert all(isinstance(group, Group) for group in groups)


@responses.activate
def test_groups_with_params(admin, get_groups_as_admin_with_expand):
    params = {
        "$expand": "dashboards",
        "$top": 100,
    }

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups",
        body=get_groups_as_admin_with_expand,
        match=[
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    groups = admin.groups(expand="dashboards", top=100)

    assert isinstance(groups, list)
    assert all(isinstance(group, Group) for group in groups)
    assert hasattr(groups[0], "dashboards")
    assert isinstance(groups[0].dashboards, list)


@responses.activate
def test_group_users(admin, get_group_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        body=get_group_users_as_admin,
        content_type="application/json",
    )

    users = admin.group_users("f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_restore_group(admin):
    json_params = {
        "name": "Restored Workspace",
        "emailAddress": "john@contoso.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/3bec11ee-48a9-490c-8e4d-1ebba90d491a/restore",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    admin.restore_group(
        "3bec11ee-48a9-490c-8e4d-1ebba90d491a",
        email_address="john@contoso.com",
        name="Restored Workspace",
    )


@responses.activate
def test_update_group_log_analytics_workspace_none(admin):
    json_params = {
        "logAnalyticsWorkspace": None,
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/e2284830-c8dc-416b-b19a-8cdcd2729332",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    admin.update_group(
        group="e2284830-c8dc-416b-b19a-8cdcd2729332",
        log_analytics_workspace=None,
    )


@responses.activate
def test_update_group_multi(admin):
    json_params = {
        "name": "New Group Name",
        "logAnalyticsWorkspace": None,
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/e2284830-c8dc-416b-b19a-8cdcd2729332",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    admin.update_group(
        group="e2284830-c8dc-416b-b19a-8cdcd2729332",
        name="New Group Name",
        log_analytics_workspace=None,
    )


def test_update_group_raises(admin):
    with pytest.raises(ValueError):
        admin.update_group(group="e2284830-c8dc-416b-b19a-8cdcd2729332")


@responses.activate
def test_report_subscriptions_call(admin, get_report_subscriptions_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/reports/5b218778-e7a5-4d73-8187-f10824047715/subscriptions",
        body=get_report_subscriptions_as_admin,
        content_type="application/json",
    )

    subscriptions = admin.report_subscriptions("5b218778-e7a5-4d73-8187-f10824047715")

    assert isinstance(subscriptions, list)
    assert all(isinstance(subscription, dict) for subscription in subscriptions)


@responses.activate
def test_report_users(admin, get_report_users_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/reports/5b218778-e7a5-4d73-8187-f10824047715/users",
        body=get_report_users_as_admin,
        content_type="application/json",
    )

    users = admin.report_users("5b218778-e7a5-4d73-8187-f10824047715")

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_reports(admin, get_reports_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/reports",
        body=get_reports_as_admin,
        content_type="application/json",
    )

    reports = admin.reports()

    assert isinstance(reports, list)
    assert all(isinstance(report, Report) for report in reports)

    assert reports[0].group_id == "278e22a3-2aee-4057-886d-c3225423bc10"
    assert reports[0].group_id == reports[0].workspace_id


@responses.activate
def test_reports_with_group(admin, get_reports_in_group_as_admin):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports",
        body=get_reports_in_group_as_admin,
        content_type="application/json",
    )

    reports = admin.reports(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(reports, list)
    assert all(isinstance(report, Report) for report in reports)

    assert reports[0].group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert not hasattr(reports[0], "workspace_id")


@responses.activate(registry=OrderedRegistry)
def test_activity_events(admin, get_activity_events):
    initial_response = responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/activityevents",
        body=get_activity_events[0],
        match=[
            matchers.query_param_matcher(
                {
                    "startDateTime": "'2019-08-31T00:00:00'",
                    "endDateTime": "'2019-08-31T23:59:59'",
                }
            )
        ],
        content_type="application/json",
    )

    next_response_1 = responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/activityevents",
        body=get_activity_events[1],
        match=[
            matchers.query_param_matcher(
                {
                    "continuationToken": "'%2BRID%3A244SAKlHY7YQAAAAAAAAAA%3D%3D%23RT%3A1%23TRC%3A5%23FPC%3AARAAAAAAAAAAFwAAAAAAAAA%3D'"
                }
            )
        ],
        content_type="application/json",
    )

    next_response_2 = responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/activityevents",
        body=get_activity_events[2],
        match=[
            matchers.query_param_matcher(
                {
                    "continuationToken": "'%2BRID%$4Z244SAKlHY7YQAAAAAAAAAA%3D%3D%23RT%3A1%23TRC%3A5%23FPC%3AARAAAAAAAAAAFwAAAAAAAAA%3D'"
                }
            )
        ],
        content_type="application/json",
    )

    start = datetime(2019, 8, 31, 0, 0, 0)
    end = datetime(2019, 8, 31, 23, 59, 59)

    activity_events = admin.activity_events(start, end)

    assert initial_response.call_count == 1
    assert next_response_1.call_count == 1
    assert next_response_2.call_count == 1

    assert isinstance(activity_events, list)
    assert all(isinstance(activity_event, dict) for activity_event in activity_events)
    assert len(activity_events) == 6
    assert activity_events[5]["Id"] == "1db4c464-3e5d-4a89-b412-c2ce6fbae88e"


@responses.activate
def test_workspaces(admin, get_modified_workspaces):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/modified",
        body=get_modified_workspaces,
        content_type="application/json",
    )

    workspaces = admin.workspaces()

    assert isinstance(workspaces, list)
    assert len(workspaces) == 2
    assert all(workspace.get("Id") for workspace in workspaces)
    assert workspaces[0].get("Id") == "3740504d-1f93-42f9-8e9d-c8ba9b787a3b"


@responses.activate
def test_workspaces_modified_since(admin, get_modified_workspaces):
    params = {"modifiedSince": "2020-10-02T05:51:30.0000000Z"}

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/modified",
        body=get_modified_workspaces,
        content_type="application/json",
        match=[
            matchers.query_param_matcher(params),
        ],
    )

    modified_since = datetime(2020, 10, 2, 5, 51, 30)

    workspaces = admin.workspaces(modified_since=modified_since)

    assert isinstance(workspaces, list)
    assert len(workspaces) == 2
    assert all(workspace.get("Id") for workspace in workspaces)
    assert workspaces[0].get("Id") == "3740504d-1f93-42f9-8e9d-c8ba9b787a3b"


@responses.activate
def test_workspaces_params(admin, get_modified_workspaces):
    params = params = {
        "modifiedSince": "2020-10-02T05:51:30.0000000Z",
        "excludeInActiveWorkspaces": False,
        "excludePersonalWorkspaces": True,
    }

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/modified",
        body=get_modified_workspaces,
        content_type="application/json",
        match=[
            matchers.query_param_matcher(params),
        ],
    )

    modified_since = datetime(2020, 10, 2, 5, 51, 30)

    workspaces = admin.workspaces(
        modified_since=modified_since,
        exclude_inactive_workspaces=False,
        exclude_personal_workspaces=True,
    )

    assert isinstance(workspaces, list)
    assert len(workspaces) == 2
    assert all(workspace.get("Id") for workspace in workspaces)
    assert workspaces[0].get("Id") == "3740504d-1f93-42f9-8e9d-c8ba9b787a3b"


@responses.activate
def test_initiate_scan(admin, post_workspace_info):
    json_params = {
        "workspaces": [
            "97d03602-4873-4760-b37e-1563ef5358e3",
        ]
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/getInfo",
        body=post_workspace_info,
        match=[
            matchers.json_params_matcher(json_params),
        ],
        content_type="application/json",
    )

    workspaces = "97d03602-4873-4760-b37e-1563ef5358e3"

    scan_request = admin.initiate_scan(workspaces)

    assert isinstance(scan_request, dict)
    assert scan_request["id"] == "e7d03602-4873-4760-b37e-1563ef5358e3"
    assert scan_request["createdDateTime"] == "2020-06-15T16:46:28.0487687Z"
    assert scan_request["status"] == "NotStarted"


@responses.activate
def test_initiate_scan_params(admin, post_workspace_info):
    json_params = {
        "workspaces": [
            "97d03602-4873-4760-b37e-1563ef5358e3",
            "67b7e93a-3fb3-493c-9e41-2c5051008f24",
        ]
    }

    params = {
        "datasetExpressions": True,
        "lineage": True,
        "datasourceDetails": False,
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/getInfo",
        body=post_workspace_info,
        match=[
            matchers.json_params_matcher(json_params),
            matchers.query_param_matcher(params),
        ],
        content_type="application/json",
    )

    workspaces = [
        "97d03602-4873-4760-b37e-1563ef5358e3",
        "67b7e93a-3fb3-493c-9e41-2c5051008f24",
    ]

    scan_request = admin.initiate_scan(
        workspaces,
        lineage=True,
        dataset_expressions=True,
        datasource_details=False,
    )

    assert isinstance(scan_request, dict)
    assert scan_request["id"] == "e7d03602-4873-4760-b37e-1563ef5358e3"
    assert scan_request["createdDateTime"] == "2020-06-15T16:46:28.0487687Z"
    assert scan_request["status"] == "NotStarted"


@responses.activate
def test_scan_status(admin, get_scan_status):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/scanStatus/e7d03602-4873-4760-b37e-1563ef5358e3",
        body=get_scan_status,
        content_type="application/json",
    )

    scan_status = admin.scan_status("e7d03602-4873-4760-b37e-1563ef5358e3")

    assert isinstance(scan_status, dict)
    assert scan_status["id"] == "e7d03602-4873-4760-b37e-1563ef5358e3"
    assert scan_status["status"] == "Succeeded"


@responses.activate
def test_scan_result(admin, get_scan_result):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/scanResult/e7d03602-4873-4760-b37e-1563ef5358e3",
        body=get_scan_result,
        content_type="application/json",
    )

    scan_result = admin.scan_result("e7d03602-4873-4760-b37e-1563ef5358e3")

    assert isinstance(scan_result, dict)
    assert len(scan_result["workspaces"]) == 1
    assert scan_result["workspaces"][0]["id"] == "d507422c-8d6d-4361-ac7a-30074a8cd0a1"
    assert all(
        k in scan_result["workspaces"][0].keys()
        for k in ("reports", "dashboards", "datasets", "dataflows")
    )



"""
Module implements a wrapper around the Power BI Rest API App operations.

Full App Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/apps

"""

from requests import Session

from pbipy.dashboards import Dashboard, Tile
from pbipy.reports import Report
from pbipy.resources import Resource
from pbipy import _utils


class App(Resource):
    """
    A Power BI installed App.
    
    Users should initialize an `App` object by calling the `app()` method
    on the `PowerBI` client, rather than creating directly.

    Examples
    --------
    Retrieving an `App` object using a `pbi` client object.
    
    ```
    >>> my_app = pbi.app("f089354e-8366-4e18-aea3-4cb4a3a50b48")
    ```

    """

    _REPR = [
        "id",
        "name",
        "description",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        raw=None,
    ) -> None:
        super().__init__(id, session)

        self.resource_path = f"/apps/{self.id}"
        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        if raw:
            self._load_from_raw(raw)

    def dashboard(
        self,
        dashboard: str,
    ) -> Dashboard:
        """
        Return the specified Dashboard from the App.

        Parameters
        ----------
        dashboard : `str`
            Dashboard Id of the Dashboard to retrieve.

        Returns
        -------
        `Dashboard`
            The specified Dashboard.

        """

        resource = self.base_path + f"/dashboards/{dashboard}"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        dashboard = Dashboard(
            id=raw.get("id"),
            session=self.session,
            raw=raw,
        )

        return dashboard

    def dashboards(
        self,
    ) -> list[Dashboard]:
        """
        Returns a list of Dashboards from the App.

        Returns
        -------
        `list[Dashboard]`
            List of Dashboards associated with the App.

        """

        resource = self.base_path + "/dashboards"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        dashboards = [
            Dashboard(
                id=dashboard_js.get("id"),
                session=self.session,
                raw=dashboard_js,
            )
            for dashboard_js in raw
        ]

        return dashboards

    def report(
        self,
        report: str,
    ) -> Report:
        """
        Return the specified report from the specified app.

        Parameters
        ----------
        `report` : `str`
            The Report Id

        Returns
        -------
        `Report`
            A Power BI report. The API returns a subset of the following
            list of report properties. The subset depends on the API called,
            caller permissions, and the availability of data in the Power
            BI database.

        """

        resource = self.base_path + f"/reports/{report}"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        report = Report(
            id=raw.get("id"),
            session=self.session,
            raw=raw,
        )

        return report

    def reports(
        self,
    ) -> list[Report]:
        """
        Returns a list of reports from the app.

        Returns
        -------
        `list[Report]`
            The collection of reports from the app.

        """

        resource = self.base_path + "/reports"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        reports = [
            Report(
                report_js.get("id"),
                self.session,
                raw=report_js,
            )
            for report_js in raw
        ]

        return reports

    def tile(
        self,
        tile: str,
        dashboard: str,
    ) -> Tile:
        """
        Return the specified tile from the specified dashboard for the App.

        Parameters
        ----------
        `tile` : `str`
            Tile Id of the tile to retrieve.
        `dashboard` : `str`
            Dashboard Id of the dashboard to retrieve.

        Returns
        -------
        `Tile`
            The specified tile.

        """

        resource = self.base_path + f"/dashboards/{dashboard}/tiles/{tile}"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        tile = Tile(
            tile,
            dashboard,
            session=self.session,
            raw=raw,
        )

        return tile

    def tiles(
        self,
        dashboard: str,
    ) -> list[Tile]:
        """
        Returns a list of `Tile` objects within the specified dashboard
        from the App.

        Parameters
        ----------
        `dashboard` : `str`
            Dashboard Id of the Dashboard that includes the tiles.

        Returns
        -------
        `list[Tile]`
            List of `Tile` objects from the specified dashboard and App.

        """

        resource = self.base_path + f"/dashboards/{dashboard}/tiles"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        tiles = [
            Tile(
                tile_js.get("id"),
                dashboard_id=dashboard,
                session=self.session,
                raw=tile_js,
            )
            for tile_js in raw
        ]

        return tiles



import pytest
import responses
from responses import matchers
import requests

from pbipy.groups import Group


@pytest.fixture
def group():
    raw = {
        "id": "e2284830-c8dc-416b-b19a-8cdcd2729332",
        "isReadOnly": False,
        "isOnDedicatedCapacity": False,
        "name": "sample group",
    }

    group = Group(
        raw.get("id"),
        session=requests.Session(),
        raw=raw,
    )

    return group


def test_group_creation_from_raw():
    raw = {
        "id": "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        "isReadOnly": False,
        "isOnDedicatedCapacity": False,
        "name": "sample group",
    }

    group = Group(raw.get("id"), session=requests.Session(), raw=raw)

    assert group.id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert group.is_read_only == False
    assert group.is_on_dedicated_capacity == False
    assert group.name == "sample group"


@responses.activate
def test_groups_call(powerbi, get_groups):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups",
        body=get_groups,
        content_type="application/json",
    )

    powerbi.groups()


@responses.activate
def test_group_add_user_call():
    json_params = {
        "identifier": "1f69e798-5852-4fdd-ab01-33bb14b6e934",
        "groupUserAccessRight": "Admin",
        "principalType": "App",
        "emailAddress": "john@contoso.com",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    group.add_user(
        identifier="1f69e798-5852-4fdd-ab01-33bb14b6e934",
        principal_type="App",
        access_right="Admin",
        email_address="john@contoso.com",
    )


@responses.activate
def test_group_remove_user_call():
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users/john@contoso.com",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    group.delete_user("john@contoso.com")


@responses.activate
def test_group_update_user_call():
    json_params = {
        "identifier": "1f69e798-5852-4fdd-ab01-33bb14b6e934",
        "groupUserAccessRight": "Admin",
        "principalType": "App",
        "emailAddress": "john@contoso.com",
    }

    responses.put(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    group.update_user(
        identifier="1f69e798-5852-4fdd-ab01-33bb14b6e934",
        principal_type="App",
        access_right="Admin",
        email_address="john@contoso.com",
    )


@responses.activate
def test_group_users_call(get_group_users):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        body=get_group_users,
        content_type="application/json",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    group.users()


@responses.activate
def test_group_users_call_params(get_group_users):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users?$top=3&$skip=1",
        body=get_group_users,
        content_type="application/json",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    group.users(top=3, skip=1)


@responses.activate
def test_group_users_result(get_group_users):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/users",
        body=get_group_users,
        content_type="application/json",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    users = group.users()

    assert isinstance(users, list)
    assert all(isinstance(user, dict) for user in users)


@responses.activate
def test_create_group_call(powerbi, create_group):
    json_params = {"name": "sample workspace"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=create_group,
        content_type="application/json",
    )

    powerbi.create_group("sample workspace")


@responses.activate
def test_create_group_call_with_workspacev2(powerbi, create_group):
    json_params = {"name": "sample workspace"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups?workspaceV2=True",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=create_group,
        content_type="application/json",
    )

    powerbi.create_group("sample workspace", workspace_v2=True)


@responses.activate
def test_create_group_result(powerbi, create_group):
    json_params = {"name": "sample workspace"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body=create_group,
        content_type="application/json",
    )

    group = powerbi.create_group("sample workspace")

    assert isinstance(group, Group)
    assert hasattr(group, "id")
    assert hasattr(group, "name")
    assert hasattr(group, "is_on_dedicated_capacity")

    assert group.id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert not group.is_on_dedicated_capacity
    assert group.name == "sample group"


@responses.activate
def test_group_update(group):
    json_params = {
        "name": "Sample Group 1",
        "defaultDatasetStorageFormat": "Small",
    }
    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/e2284830-c8dc-416b-b19a-8cdcd2729332",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    group.update(name="Sample Group 1", default_dataset_storage_format="Small")

    # Changes were made
    assert group.name == "Sample Group 1"
    assert group.default_dataset_storage_format == "Small"
    assert group.raw.get("defaultDatasetStorageFormat") == "Small"

    # Existing values unaffected
    assert group.is_read_only == False
    assert group.raw.get("isOnDedicatedCapacity") == False


@responses.activate
def test_group_update_name(group):
    json_params = {
        "name": "Sample Group 1",
    }
    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/e2284830-c8dc-416b-b19a-8cdcd2729332",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    group.update(name="Sample Group 1")

    assert group.name == "Sample Group 1"
    assert group.raw.get("name") == "Sample Group 1"


@responses.activate
def test_group_update_raises(group):
    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/e2284830-c8dc-416b-b19a-8cdcd2729332",
    )

    with pytest.raises(ValueError):
        group.update()



import pytest
import requests

from pbipy.dashboards import Dashboard, Tile


@pytest.fixture
def dashboard_with_group():
    raw = {
        "id": "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        "displayName": "SalesMarketing",
        "embedUrl": "https://app.powerbi.com/dashboardEmbed?dashboardId=69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        "isReadOnly": False,
    }

    dashboard = Dashboard(
        id=raw.get("id"),
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
        raw=raw,
    )

    return dashboard


@pytest.fixture
def tile():
    raw = {
        "id": "312fbfe9-2eda-44e0-9ed0-ab5dc571bb4b",
        "title": "SalesMarketingTile",
        "embedUrl": "https://app.powerbi.com/embed?dashboardId=69ffaa6c-b36d-4d01-96f5-1ed67c64d4af&tileId=312fbfe9-2eda-44e0-9ed0-ab5dc571bb4b&groupId=f089354e-8366-4e18-aea3-4cb4a3a50b48",
        "rowSpan": 0,
        "colSpan": 0,
        "reportId": "5b218778-e7a5-4d73-8187-f10824047715",
        "datasetId": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
    }

    tile = Tile(
        id=raw.get("id"),
        dashboard_id="69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        session=requests.Session(),
        raw=raw,
    )

    return tile


def test_dashboard_creation(dashboard_with_group):
    assert isinstance(dashboard_with_group, Dashboard)
    assert dashboard_with_group.id == "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af"


def test_tile_creation(tile):
    assert isinstance(tile, Tile)
    assert tile.row_span == 0
    assert tile.col_span == 0
    assert tile.report_id == "5b218778-e7a5-4d73-8187-f10824047715"
    assert tile.title == "SalesMarketingTile"



from types import SimpleNamespace

from requests import Session

from pbipy.groups import Group
from pbipy.resources import Resource


class Import(Resource):
    """
    A file that has been uploaded into Power BI.

    """

    _REPR = [
        "id",
        "name",
        "group_id",
        "created_date_time",
        "updated_date_time",
        "import_state",
        "connection_type",
        "source",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        group_id: str | Group = None,
        raw: dict = None,
    ) -> None:
        super().__init__(id, session)

        if group_id:
            self.group_id = group_id
        else:
            self.group_id = None

        if self.group_id:
            self.resource_path = f"/groups/{self.group_id}/imports/{self.id}"
        else:
            self.resource_path = f"/imports/{self.id}"

        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        if raw:
            self._load_from_raw(raw)


class TemporaryUploadLocation(SimpleNamespace):
    """
    Temporary blob storage location for importing large Power BI `.pbix`
    files.

    Parameters
    ----------
    `expiration_time` : `str`
        The expiration date and time of the shared access signature URL.
    `url` : `str`
        The shared access signature URL for the temporary blob storage.

    """

    def __init__(
        self,
        expiration_time: str,
        url: str,
    ) -> None:
        self.expiration_time = expiration_time
        self.url = url



"""
Module implements a wrapper around the Power BI Rest API Dataset operations.

Full Dataset Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/datasets

"""

import time

from requests import Session

from pbipy.resources import Resource
from pbipy import _utils


# TODO: Move to an exceptions.py module
class DatasetRefreshError(Exception):
    """Error raised when a Dataset refresh did not complete successfully."""

    def __init__(
        self,
        refresh_id: str,
        status: str,
        refresh_details: dict,
    ):
        self.refresh_id = refresh_id
        self.status = status
        self.refresh_details = refresh_details

        message = f"Refresh {self.refresh_id} did not complete successfully. Status: {self.status}. See the 'refresh_details' property for more information."

        super().__init__(message)


class Dataset(Resource):
    """
    A Power BI Dataset.

    Users should initialize a `Dataset` object by calling the `dataset()`
    method on the `PowerBI` client, rather than creating directly.

    Examples
    --------
    Retrieving a `Dataset` object using a `pbi` client object.

    ```
    >>> my_dataset = pbi.dataset("cfafbeb1-8037-4d0c-896e-a46fb27ff229")
    ```

    """

    _REPR = [
        "id",
        "name",
        "group_id",
        "configured_by",
        "created_date",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        group_id=None,
        raw=None,
    ) -> None:
        super().__init__(id, session)

        if group_id:
            self.group_id = group_id
        else:
            self.group_id = None

        if self.group_id:
            self.resource_path = f"/groups/{self.group_id}/datasets/{self.id}"
        else:
            self.resource_path = f"/datasets/{self.id}"

        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        # Supports creating from a list of js, e.g., get_datasets endpoint
        if raw:
            self._load_from_raw(raw)

    def add_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
    ) -> None:
        """
        Grants the specified user's permissions to the specified dataset.

        Parameters
        ----------
        `identifier` : `str`
            For principal type `User`, provide the UPN. Otherwise provide
            the object ID of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The Dataset User Access Right to grant to the user for the
            dataset, e.g.,"Read", "ReadExplore", "ReadReshare", or
            "ReadReshareExplore".

        """

        resource = self.base_path + "/users"
        dataset_user_access = {
            "identifier": identifier,
            "principalType": principal_type,
            "datasetUserAccessRight": access_right,
        }

        _utils.post(
            resource,
            self.session,
            dataset_user_access,
        )

    def bind_to_gateway(
        self,
        gateway_object_id: str,
        datasource_object_ids: list = None,
    ) -> None:
        """
        Binds the specified dataset from MyWorkspace or group to the specified
        gateway, optionally with a given set of data source IDs. If you
        don't supply a specific data source ID, the dataset will be bound
        to the first matching data source in the gateway.

        Parameters
        ----------
        `gateway_object_id` : `str`
            The gateway ID. When using a gateway cluster, the gateway ID
            refers to the primary (first) gateway in the cluster and is
            similar to the gateway cluster ID.
        `datasource_object_ids` : `list`, optional
            The unique identifiers for the data sources in the gateway.

        """

        bind_to_gateway_request = {
            "gatewayObjectId": gateway_object_id,
            "datasourceObjectIds": datasource_object_ids,
        }

        if bind_to_gateway_request["datasourceObjectIds"] is None:
            bind_to_gateway_request.pop("datasourceObjectIds")

        resource = self.base_path + "/Default.BindToGateway"

        _utils.post(
            resource,
            self.session,
            bind_to_gateway_request,
        )

    def cancel_refresh(
        self,
        refresh_id: str,
    ) -> None:
        """
        Cancels the specified refresh operation for the specified dataset
        from MyWorkspace or group.

        Parameters
        ----------
        `refresh_id` : `str`
            Refresh Id to cancel.

        """

        resource = self.base_path + f"/refreshes/{refresh_id}"
        _utils.delete(
            resource,
            self.session,
        )

    def datasources(
        self,
    ) -> list[dict]:
        """
        Returns a list of datasources for the dataset.

        Returns
        -------
        `list[dict]`
            List of PowerBI datasources for the dataset.

        """

        resource = self.base_path + "/datasources"
        return _utils.get_raw(
            resource,
            self.session,
        )

    def discover_gateways(
        self,
    ) -> list:
        """
        Returns a list of gateways that the dataset can be bound to.

        This API call is only relevant to datasets that have at least one
        on-premises connection. For datasets with cloud-only connections,
        this API call returns an empty list.

        Returns
        -------
        `list`
            List of PowerBI Gateways that can be bound to.

        """

        resource = self.base_path + "/Default.DiscoverGateways"
        return _utils.get_raw(
            resource,
            self.session,
        )

    def execute_queries(
        self,
        queries: str | list[str],
        impersonated_user_name: str = None,
        include_nulls: bool = None,
    ) -> dict:
        """
        Executes Data Analysis Expressions (DAX) queries against the provided
        dataset.

        DAX query errors will result in:
            A response error, such as DAX query failure.
            A failure HTTP status code (400).

        A query that requests more than one table, or more than the allowed
        number of table rows, will result in:
            Limited data being returned.
            A response error, such as More than one result table in a query
            or More than {allowed number} rows in a query result.
            A successful HTTP status code (200).

        Columns that are fully qualified in the query will be returned with
        a fully qualified name, for example, MyTable[MyColumn]. Columns
        that are renamed or created in the query will be returned within
        square bracket, for example, `[MyNewColumn]`.

        Parameters
        ----------
        `queries` : `str | list[str]`
            Query or list of queries to execute against the dataset.
        `impersonated_user_name` : `str`, optional
            The UPN of a user to be impersonated. If the model is not RLS
            enabled, this will be ignored.
        `include_nulls` : `bool`, optional
            Whether null (blank) values should be included in the result
            set. If unspecified, the default value is `false`.

        Returns
        -------
        `dict`
            Dict containing the results of the execution.

        """

        if isinstance(queries, str):
            qs = [{"query": queries}]
        else:
            qs = [{"query": query} for query in queries]

        dataset_execute_queries_request = {
            "queries": qs,
            "serializerSettings": {
                "includeNulls": include_nulls,
            },
            "impersonatedUserName": impersonated_user_name,
        }
        prepared_request = _utils.remove_no_values(dataset_execute_queries_request)

        resource = self.base_path + "/executeQueries"

        raw = _utils.post_raw(
            resource,
            self.session,
            payload=prepared_request,
        )

        return raw

    def parameters(
        self,
    ) -> list[dict]:
        """
        Return a list of parameters for the dataset.

        Returns
        -------
        `list[dict]`
            Parameter list.

        """

        resource = self.base_path + "/parameters"
        return _utils.get_raw(
            resource,
            self.session,
        )

    def refresh(
        self,
        notify_option: str = None,
        apply_refresh_policy: bool = None,
        commit_mode: str = None,
        effective_date: str = None,
        max_parallelism: int = None,
        objects: list[dict] = None,
        retry_count: int = None,
        type: str = None,
    ) -> str:
        """
        Triggers a refresh, or enhanced refresh, of the Dataset and returns
        the generated Refresh Id.

        If no parameters, or only `notify_option`, are provided then a
        standard refresh is triggered.

        To trigger an enhanced refresh, provide one or more of the optional
        parameters other than `notify_option`. When the enhanced refresh
        specifies parameters, the API sets unspecified parameters to their
        default values.

        Parameters
        ----------
        `notify_option` : `str`, optional
            Mail notification options, e.g., "MailOnCompletion", "MailOnFailure",
            or "NoNotification".
        `apply_refresh_policy` : `bool`, optional
            Determine if the policy is applied or not.
        `commit_mode` : `str`, optional
            Determines if objects will be committed in batches or only when
            complete, e.g., "PartialBatch", or "Transactional".
        `effective_date` : `str`, optional
            If an incremental refresh policy is applied, the `effective_date`
            parameter overrides the current date.
        `max_parallelism` : `int`, optional
            The maximum number of threads on which to run parallel processing
            commands.
        `objects` : `list[dict]`, optional
            A list of objects to be processed, e.g.,
            ```
            [
                {
                "table": "Customer",
                "partition": "Robert"
                }
            ]
            ```
        `retry_count` : `int`, optional
            Number of times the operation will retry before failing.
        `type` : `str`, optional
           The type of processing to perform, e.g., "Automatic", "Calculate",
           "ClearValues", "DataOnly", "Defragment", or "Full".

        Returns
        -------
        `str`
            Refresh Id that was created by triggering the refresh. If the
            triggered refresh was an enhanced refresh, the Refresh Id can
            be passed to `Dataset.refresh_details` to retrieve execution
            details for the refresh.

        Notes
        -----
        See here for request options in greater detail:
        https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/refresh-dataset#definitions

        Enhanced refresh with the Power BI REST API:
        https://learn.microsoft.com/en-us/power-bi/connect-data/asynchronous-refresh

        """

        refresh_request = {
            "applyRefreshPolicy": apply_refresh_policy,
            "commitMode": commit_mode,
            "effectiveDate": effective_date,
            "maxParallelism": max_parallelism,
            "notifyOption": notify_option,
            "objects": objects,
            "retryCount": retry_count,
            "type": type,
        }

        prepared_request = _utils.remove_no_values(refresh_request)
        resource = self.base_path + "/refreshes"

        response = _utils.post(
            resource,
            self.session,
            prepared_request,
        )

        return response.headers["RequestId"]

    def refresh_details(
        self,
        refresh_id: str,
    ) -> dict:
        """
        Returns execution details of an enhanced refresh operation for
        the dataset.

        Parameters
        ----------
        `refresh_id` : `str`
            Refresh Id to get the execution details for.

        Returns
        -------
        `dict`
            Refresh execution details.

        """

        resource = self.base_path + f"/refreshes/{refresh_id}"
        return _utils.get_raw(
            resource,
            self.session,
        )

    def refresh_and_wait(
        self,
        apply_refresh_policy: bool = None,
        commit_mode: str = None,
        effective_date: str = None,
        max_parallelism: int = None,
        objects: list[dict] = None,
        retry_count: int = None,
        type: str = None,
        check_interval: int = 30,
    ) -> None:
        """
        Triggers an enhanced refresh of the dataset and periodically checks
        the refresh status until the refresh operation completes. If the
        refresh completes successfully, control is returned to the caller.
        If the refresh did not complete successfully an exception is raised.

        Convenience method that triggers an enhanced refresh using `Dataset.refresh`,
        and then periodically checks the status of the refresh using `Dataset.refreh_details`.
        The frequency of checking is set via the `check_interval` parameter.

        Parameters
        ----------
        `apply_refresh_policy` : `bool`, optional
            Determine if the policy is applied or not.
        `commit_mode` : `str`, optional
            Determines if objects will be committed in batches or only when
            complete, e.g., "PartialBatch", or "Transactional".
        `effective_date` : `str`, optional
            If an incremental refresh policy is applied, the `effective_date`
            parameter overrides the current date.
        `max_parallelism` : `int`, optional
            The maximum number of threads on which to run parallel processing
            commands.
        `objects` : `list[dict]`, optional
            A list of objects to be processed, e.g.,
            ```
            [
                {
                "table": "Customer",
                "partition": "Robert"
                }
            ]
            ```
        `retry_count` : `int`, optional
            Number of times the operation will retry before failing.
        `type` : `str`, optional
           The type of processing to perform, e.g., "Automatic", "Calculate",
           "ClearValues", "DataOnly", "Defragment", or "Full".
        `check_interval` : `int`
            How often, in seconds, to check the status of the triggered
            refresh.

        Raises
        ------
        `ValueError`
            If no options were provided for the refresh. At least one of the
            refresh options must be provided to the endpoint to trigger an enhanced
            refresh.
        `DatasetRefreshError`
            When the refresh status finished with an unsuccessful status.

        Notes
        -----
        An enhanced refresh is triggered when additional parameters, other
        than "notifyOption" are provided to the Refresh Dataset endpoint.
        Due to the Refresh Execution Details endpoint only supporting enhanced
        refreshes, this method also only supports enhanced refreshes.

        Enhanced refresh with the Power BI REST API:
        https://learn.microsoft.com/en-us/power-bi/connect-data/asynchronous-refresh

        Refresh Dataset endpoint:
        https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/refresh-dataset

        """

        if not any(
            [
                apply_refresh_policy,
                commit_mode,
                effective_date,
                max_parallelism,
                objects,
                retry_count,
                type,
            ]
        ):
            raise ValueError(
                "No options were provided for the refresh. Must provide at least one of: apply_refresh_policy, commit_mode, effective_date, max_parallelism, objects, or retry_count."
            )

        refresh_id = self.refresh(
            apply_refresh_policy=apply_refresh_policy,
            commit_mode=commit_mode,
            effective_date=effective_date,
            max_parallelism=max_parallelism,
            objects=objects,
            retry_count=retry_count,
            type=type,
        )

        refresh_details = self.refresh_details(refresh_id)
        status = refresh_details.get("status", "Unknown")

        finished = ("Completed", "Failed", "Disabled", "Cancelled")

        while status not in finished:
            time.sleep(check_interval)

            refresh_details = self.refresh_details(refresh_id)
            status = refresh_details.get("status", "Unknown")

        if status != "Completed":
            raise DatasetRefreshError(refresh_id, status, refresh_details)

    def refresh_history(
        self,
        top: int = None,
    ) -> list[dict]:
        """
        Returns the refresh history for the dataset.

        Parameters
        ----------
        `top` : `int`, optional
            The requested number of entries in the refresh history. If
            not provided, the default is the last available 500 entries.

        Returns
        -------
        `list[dict]`
            List of refresh history entries.

        Raises
        ------
        `HTTPError`
            If the api response status code is not equal to 200.

        """
        # TODO: implement Refresh object

        resource = self.base_path + "/refreshes"
        params = {"$top": top}

        raw = _utils.get_raw(
            resource,
            self.session,
            params,
        )

        return raw

    def refresh_schedule(
        self,
        direct_query: bool = False,
    ) -> dict:
        """
        Return the Refresh Schedule or Direct Query Refresh Schedule for
        the dataset.

        Parameters
        ----------
        `direct_query` : `bool`, optional
            Return the Direct Query Refresh Schedule instead of the Refresh Schedule.

        Returns
        -------
        `dict`
            Refresh schedule or Direct Query Refresh Schedule.

        """

        if direct_query:
            resource = self.base_path + "/directQueryRefreshSchedule"
        else:
            resource = self.base_path + "/refreshSchedule"

        raw = _utils.get_raw(
            resource,
            self.session,
        )

        return raw

    def take_over(
        self,
    ) -> None:
        """
        Transfer ownership of the dataset to the current authorized user.

        Raises
        ------
        `TypeError`
            If the dataset does not have a group_id. In other words, can't
            take over dataset in MyWorkspace, the authorized user already
            owns these.

        """

        if not self.group_id:
            raise TypeError(
                "Dataset does not have a group_id. Taking over a dataset can only be performed on a Dataset in a Group."
            )

        resource = self.base_path + "/Default.TakeOver"
        _utils.post(
            resource,
            self.session,
        )

    def update(
        self,
        target_storage_mode: str,
    ) -> None:
        """
        Update the properties of the dataset.

        Parameters
        ----------
        `target_storage_mode` : `str`
            The dataset storage mode, .e.g, "PremiumFiles", or "Abf".

        """
        update_dataset_request = {"targetStorageMode": target_storage_mode}

        _utils.patch(
            self.base_path,
            self.session,
            update_dataset_request,
        )

    def update_datasources(
        self,
        update_details: dict | list[dict],
    ) -> None:
        """
        Update the data sources of the dataset.

        Parameters
        ----------
        `update_details` : `dict | list[dict]`
            A dict, or list of dicts, representing the updates.

        Notes
        -----
        See here for how to construct the update details:
        https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/update-datasources#examples

        """

        if isinstance(update_details, dict):
            update_details_prepared = [update_details]
        else:
            update_details_prepared = update_details

        update_request = {"updateDetails": update_details_prepared}

        resource = self.base_path + "/Default.UpdateDatasources"
        _utils.post(
            resource,
            self.session,
            update_request,
        )

    def update_parameters(
        self,
        update_details: dict | list[dict],
    ) -> None:
        """
        Updates the parameters values for the dataset.

        Parameters
        ----------
        update_details : `dict | list[dict]`
            Dict, or list of dicts, of the parameters to update.
            Parameter updates take the form:
            ```
            {
                "name": "parameter_name",
                "newValue": "new_value"
            }
            ```

        Notes
        -----
        See more here:
        https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/update-parameters#example

        """

        if isinstance(update_details, dict):
            update_details_prepared = [update_details]
        else:
            update_details_prepared = update_details

        request_body = {
            "updateDetails": update_details_prepared,
        }

        resource = self.base_path + "/Default.UpdateParameters"
        _utils.post(
            resource,
            self.session,
            request_body,
        )

    def update_refresh_schedule(
        self,
        notify_option: str = None,
        direct_query: bool = False,
        days: list[str] = None,
        enabled: bool = None,
        frequency: int = None,
        local_time_zone_id: str = None,
        times: list[str] = None,
    ) -> None:
        """
        Update the Refresh Schedule or Direct Query Refresh Schedule of the
        dataset.

        Providing `direct_query=True` targets the Direct Query
        Refresh Schedule.

        Parameters
        ----------
        `notify_option` : `str`
            Email notification options, e.g., "MailOnCompletion", "MailOnFailure",
            or "NoNotification".
        `direct_query` : bool, optional
            Target the Direct Query Refresh Schedule (if there is one) of the
            dataset.
        `days` : list[str], optional
            The full name of days on which to execute the refresh,e.g, "Monday",
            "Tuesday", "Wednesday", etc.
        `enabled` : bool, optional
            Whether the refresh is enabled.
        `frequency` : int, optional
            Applies to Direct Query Refresh Schedule only. The interval in minutes
            between successive refreshes. Supported values are 15, 30, 60, 120,
            and 180.
            If `frequency` is supplied and `direct_query` is `false`, then `frequency`
            will be ignored.
        `local_time_zone_id` : `str`, optional
            The ID of the time zone to use, e.g, "UTC".
        `times` : list[str], optional
            The times of day to execute the refresh expressed as hh:mm, e.g.,
            "07:00", "16:00", etc.

        Raises
        ------
        `ValueError`
            If no values are provided for the request.

        """

        # Prepare the request details
        refresh_schedule_request = {
            "value": {
                "notifyOption": notify_option,
                "days": days,
                "enabled": enabled,
                "localTimeZoneId": local_time_zone_id,
                "times": times,
                # direct query option
                "frequency": frequency,
            }
        }

        if not direct_query:
            refresh_schedule_request.pop("frequency", None)

        request_body = _utils.remove_no_values(refresh_schedule_request)

        if request_body in [None, {}]:
            raise ValueError(
                "No options were provided to update. Please specify an option to update."
            )

        # Check which refresh schedule we're updating
        if direct_query:
            resource = self.base_path + "/directQueryRefreshSchedule"
        else:
            resource = self.base_path + "/refreshSchedule"

        _utils.patch(
            resource,
            self.session,
            request_body,
        )

    def update_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
    ) -> None:
        """
        Updates the existing permissions for a user of the dataset to the
        specified permissions.

        Parameters
        ----------
        `identifier` : `str`
            For principal type User, provide the UPN. Otherwise provide
            the object ID of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The Dataset User Access Right to grant to the user, e.g.,"Read",
            "ReadExplore", "ReadReshare", or "ReadReshareExplore".

        """

        resource = self.base_path + "/users"
        dataset_user_access = {
            "identifier": identifier,
            "principalType": principal_type,
            "datasetUserAccessRight": access_right,
        }

        _utils.put(
            resource,
            self.session,
            dataset_user_access,
        )

    def users(
        self,
    ) -> list[dict]:
        """
        Returns a list of principals that have access to the dataset.

        Returns
        -------
        `list[dict]`
            List of principals, e.g., Users, Groups, with access to the
            dataset.

        """

        resource = self.base_path + "/users"
        return _utils.get_raw(
            resource,
            self.session,
        )



"""pbipy Resource definition."""

from requests import Session

from pbipy import settings
from pbipy import _utils


class Resource:
    """
    Represents a URL-addressable resource in the Power BI Rest API.

    Resources are Power BI components such as Apps, Reports, Groups, 
    Datasets, etc. and closely mirror the REST Operation Groups here:

    https://learn.microsoft.com/en-us/rest/api/power-bi/
    
    Resources translate the JSON returned by the API into more user-friendly 
    Python representations.

    """

    BASE_URL = settings.BASE_URL

    def __init__(
        self,
        id: str,
        session: Session,
        **kwargs,
    ) -> None:
        self.id = id
        self.session = session
        self.raw = None

    def __repr__(
        self,
    ) -> str:
        """
        Provide a readable representation of the class. Looks to the class'
        `self._REPR` to determine what to show.

        Returns
        -------
        `str`
            Formatted, readable representation of the class.
        """

        name = self.__class__.__name__

        attrs = []
        for attr in self._REPR:
            if attr in self.__dict__.keys():
                attr_str = "{}={!r}".format(attr, self.__dict__.get(attr))
                attrs.append(attr_str)

        return f"<{name} {', '.join(attrs)}>"

    def _load_from_raw(self, raw):
        self.raw = raw

        for k, v in raw.items():
            attr = _utils.to_identifier(k)
            attr = _utils.to_snake_case(attr)
            setattr(self, attr, v)

        return self

    def load(self):
        raw = _utils.get_raw(
            self.base_path,
            self.session,
        )
        self._load_from_raw(raw)



"""
Module implements a wrapper around the Power BI Rest API Report operations.

Full Report Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/reports

"""

import mimetypes
from pathlib import Path

from requests import Session
import gzip
from urllib.request import Request, urlopen

from pbipy.datasets import Dataset
from pbipy.embedtokens import EmbedToken
from pbipy.groups import Group
from pbipy.resources import Resource
from pbipy import _utils


class Report(Resource):
    """
    A Power BI Report.

    Users should initialize a `Report` object by calling the `report()`
    method on the `PowerBI` client, rather than creating directly.

    Examples
    --------
    Retrieving a `Report` object using a `pbi` client object.

    ```
    >>> my_report = pbi.report("5b218778-e7a5-4d73-8187-f10824047715")
    ```

    Retrieve the list of datasources for a report.

    ```
    >>> my_report = pbi.report("5b218778-e7a5-4d73-8187-f10824047715")
    >>> my_report.datasources()
    ```

    """

    _REPR = [
        "id",
        "name",
        "group_id",
        "dataset_id",
    ]

    REPORT_EXTENSIONS = {
        "PowerBIReport": "pbix",
        "PaginatedReport": "rdl",
    }

    def __init__(
        self,
        id: str,
        session: Session,
        group_id: str = None,
        raw: dict = None,
    ) -> None:
        super().__init__(id, session)

        if group_id:
            self.group_id = group_id
        else:
            self.group_id = None

        if self.group_id:
            self.resource_path = f"/groups/{self.group_id}/reports/{self.id}"
        else:
            self.resource_path = f"/reports/{self.id}"

        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        if raw:
            self._load_from_raw(raw)

    def clone(
        self,
        name: str,
        target_group: str = None,
        target_dataset: str = None,
    ) -> None:
        """
        Clones the report into the current user's workspace or the specified
        workspace.

        If the dataset for a cloned report resides in two different workspaces
        or in the current user's workspace, then a shared dataset will be
        created in the report's workspace.

        When cloned, reports with a live connection will lose that connection
        and instead have a direct binding to the target dataset.

        Parameters
        ----------
        `name` : `str`
            The cloned report name.
        `target_group` : `str`, optional
            Group Id specifying the group where the clone will reside.
            An empty GUID (`00000000-0000-0000-0000-000000000000`) indicates
            the current user's workspace. If this parameter isn't provided,
            the new report will be cloned within the same workspace as the
            source report.
        `target_dataset` : `str`, optional
            Dataset Id specifying the dataset to associate with the cloned
            report. If not provided, the new report will be associated with
            the same dataset as the source report.

        """

        init_payload = {
            "name": name,
            "targetModelId": target_dataset,
            "targetWorkspaceId": target_group,
        }

        payload = _utils.remove_no_values(init_payload)
        resource = self.base_path + "/Clone"

        _utils.post(resource, self.session, payload)

    def datasources(
        self,
    ) -> list[dict]:
        """
        Return a list of Datasources for the report.

        Returns
        -------
        `list[dict]`
            List of Datasources for the report.

        """

        resource = self.base_path + "/datasources"
        raw = _utils.get_raw(resource, self.session)

        return raw

    def download(
        self,
        save_to: str | Path = None,
        file_name: str = None,
    ) -> None:
        """
        Download the report as `.pbix` or `.rdl` depending on the report
        type.

        Parameters
        ----------
        `save_to` : `str | Path`, optional
            Folder/directory to save the report to. If not provided will
            save to the current working directory.
        `file_name` : `str`, optional
            Name of the file. If not provided will use the name of the
            Report as the file name.

        """

        if file_name is None:
            f_name = self.name
        else:
            f_name = file_name

        ext = self.REPORT_EXTENSIONS.get(self.report_type)

        file_path = _utils.file_path_from_components(
            file_name=f_name,
            extension=ext,
            directory=save_to,
        )

        resource = self.base_path + "/Export"
        request = Request(resource, headers=self.session.headers)
        with (
            urlopen(request) as response,
            gzip.GzipFile(fileobj=response, mode="rb") as uncompressed,
            open(file_path, "wb") as out_file,
        ):
            while True:
                chunk = uncompressed.read(1024 * 1024)
                if not chunk:
                    break
                out_file.write(chunk)

    def download_export(
        self,
        id,
        save_to: str | Path = None,
        file_name: str = None,
    ) -> None:
        """
        Download the file for a Report's export request.

        Parameters
        ----------
        `id` : `str`
            The Export Request Id to retrieve the file for.
        `save_to` : `str | Path`, optional
            Folder/directory to save the exported file to. If not provided
            will save to the current working directory.
        `file_name` : `str`, optional
            Name to give to the export file. If not provided, will use the
            name of the Report as the file name.

        """

        resource = self.base_path + f"/exports/{id}/file"
        response = _utils.get(resource, self.session)

        if file_name is None:
            f_name = self.name
        else:
            f_name = file_name

        content_type = response.headers.get("content-type")

        # Try and ensure more consistent extension guessing
        mimetypes.add_type(
            "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            ".pptx",
        )
        ext = mimetypes.guess_extension(content_type)

        file_path = _utils.file_path_from_components(
            f_name,
            extension=ext,
            directory=save_to,
        )

        with open(file_path, "wb") as out_file:
            out_file.write(response.content)

    def export_request(
        self,
        format: str,
    ) -> dict:
        """
        Trigger an export job for the report.

        Parameters
        ----------
        `format` : `str`
            The format to export to, e.g., "pdf", "png", "pptx", "xlsx".

        Returns
        -------
        `dict`
            Details and current state of the export job.

        """

        resource = self.base_path + "/ExportTo"
        payload = {"format": format.upper()}

        raw = _utils.post_raw(resource, self.session, payload)

        return raw

    def export_status(
        self,
        id: str,
        include_retry_after: bool = False,
    ) -> dict | tuple[dict, int]:
        """
        Returns the current status of the provided export request.

        Parameters
        ----------
        `id` : `str`
            Id of the export request.
        `include_retry_after` : `bool`, optional
            If included, `export_status` will inspect the Response Header for
            a Retry-After value and return this. This value can be used to delay
            the next status request, instead of polling at a set interval.

        Returns
        -------
        `dict | tuple[dict, int]`
            Dict representing the export request and its current state. If `include_retry_after`
            was set to `True`, then a tuple is returned that includes the
            export request and `retry_after` value.

        """

        resource = self.base_path + f"/exports/{id}"
        response = _utils.get(
            resource,
            self.session,
        )

        raw = _utils.parse_raw(response.json())

        if include_retry_after:
            try:
                retry_after = int(response.headers.get("Retry-After"))
            except:
                retry_after = None

            return raw, retry_after
        else:
            return raw

    def generate_token(
        self,
        access_level: str = None,
        allow_save_as: bool = None,
        dataset_id: str = None,
        identities: list[dict] = None,
        lifetime_in_minutes: int = None,
    ) -> EmbedToken:
        """
        Generates an Embed Token to view or edit the Report from the it's
        workspace.

        Parameters
        ----------
        `access_level` : `str`, optional
            The required access level. Can be one of "Create", "Edit",
            or "View".
        `allow_save_as` : `bool`, optional
            Whether an embedded report can be saved as a new report. API
            defaults to `false`. Only applies when you generate an embed
            token for report embedding.
        `dataset_id` : `str`, optional
            The dataset ID used for report creation. Only applies when you
            generate an embed token for report creation.
        `identities` : `list[dict]`, optional
            A list of identities to use for row-level security rules. See
            'Notes' below for documentation describing how to define these.
        `lifetime_in_minutes` : `int`, optional
            The maximum lifetime of the token in minutes, starting from
            the time it was generated. Can be used to shorten the expiration
            time of a token, but not to extend it. The value must be a positive
            integer. Zero (0) is equivalent to `null` and will be ignored,
            resulting in the default expiration time.

        Returns
        -------
        `EmbedToken`
            A Power BI Embed Token.

        Raises
        ------
        `TypeError`
            If the report does not have a `group_id` property, i.e., the
            report is not in a Workspace.

        Notes
        -----
        See below for API Documentation and how to define parameter options.

        https://learn.microsoft.com/en-us/rest/api/power-bi/embed-token/reports-generate-token-for-create-in-group

        """

        if not self.group_id:
            raise TypeError(
                "Report does not have a 'group_id'. Generating a token can only be performed on a Report in a Group (Workspace)."
            )

        initial_payload = {
            "accessLevel": access_level,
            "allowSaveAs": allow_save_as,
            "datasetId": dataset_id,
            "identities": identities,
            "lifetimeInMinutes": lifetime_in_minutes,
        }

        payload = _utils.remove_no_values(initial_payload)

        resource = self.base_path + "/GenerateToken"
        raw = _utils.post_raw(resource, self.session, payload)

        return EmbedToken(
            token=raw["token"],
            token_id=raw["tokenId"],
            expiration=raw["expiration"],
        )

    def page(
        self,
        name: str,
    ) -> dict:
        """
        Returns the specified page from the report.

        Parameters
        ----------
        `name` : `str`
            Name of the page to return.

        Returns
        -------
        `dict`
            The specified page.

        """

        resource = self.base_path + f"/pages/{name}"
        raw = _utils.get_raw(resource, self.session)

        return raw

    def pages(
        self,
    ) -> list[dict]:
        """
        Returns a list of pages within the report.

        Returns
        -------
        `list[dict]`
            List of report pages.

        """

        resource = self.base_path + "/pages"
        raw = _utils.get_raw(resource, self.session)

        return raw

    def rebind(
        self,
        dataset: str | Dataset,
    ) -> None:
        """
        Rebinds the report to the specified dataset.

        If the specified dataset resides in a different workspace than
        the report, then a shared dataset will be created in the report's
        workspace.

        On rebind, reports with a live connection will lose that connection
        and instead have a direct binding to the target dataset.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            The new dataset for the rebound report. If the dataset resides
            in a different workspace than the report, a shared dataset will
            be created in the report's workspace.

        """

        if isinstance(dataset, Dataset):
            dataset_id = dataset.id
        else:
            dataset_id = dataset

        payload = {
            "datasetId": dataset_id,
        }
        resource = self.base_path + "/Rebind"

        _utils.post(resource, self.session, payload)

    def take_over(
        self,
    ) -> None:
        """
        Transfers ownership of the data sources for the report to the current
        authorized user.

        Raises
        ------
        `TypeError`
            If the Report does not have a Workspace (Group), i.e., the
            Report resides in current authorized user's Workspace.

        """

        if not self.group_id:
            raise TypeError(
                "Report does not have a group_id. Taking over a Report can only be performed on a Report in a Group."
            )

        resource = self.base_path + "/Default.TakeOver"
        _utils.post(resource, self.session)

    def update_content(
        self,
        source_report: "str | Report",
        source_group: str | Group = None,
        source_type: str = "ExistingReport",
    ) -> None:
        """
        Updates the content of the report with the content of a specified
        source report.

        If the caller provides a `Report` object as a `source_report`, then
        the `group_id` of the `source_report` is used in the update request
        and any `source_group` argument is ignored.

        Parameters
        ----------
        `source_report` : `str | Report`
            The Report Id or `Report` object that is the source of the content.
        `source_group` : `str | Group`, optional
            The Group Id or `Group` object where the `source_report` resides.
            If `source_report` is a `Report` object, this argument is ignored.
        `source_type` : `str`, optional
            Type of source. API specification indicates the only valid
            value is "ExistingReport".

        """

        if isinstance(source_report, Report):
            report_id = source_report.id
        else:
            report_id = source_report

        # If we got report, use its group_id and ignore provided
        if isinstance(source_report, Report):
            group_id = source_report.group_id
        else:
            if isinstance(source_group, Group):
                group_id = source_group.id
            else:
                group_id = source_group

        payload = {
            "sourceReport": {
                "sourceReportId": report_id,
                "sourceWorkspaceId": group_id,
            },
            "sourceType": source_type,
        }

        prepared_payload = _utils.remove_no_values(payload)
        resource = self.base_path + "/UpdateReportContent"

        _utils.post(resource, self.session, prepared_payload)

    def update_datasources(
        self,
        details: dict | list[dict],
    ) -> None:
        """
        Updates the data sources of the report with supplied update details.

        Both the original data source and the new data source must have the
        exact same schema.

        Parameters
        ----------
        `details` : `dict | list[dict]`
            The update details for the data sources of the paginated report.

        Notes
        -----
        See below for the update details schema.

        https://learn.microsoft.com/en-us/rest/api/power-bi/reports/update-datasources#examples

        """

        if isinstance(details, dict):
            payload = [details]
        else:
            payload = details

        resource = self.base_path + "/Default.UpdateDatasources"

        _utils.post(resource, self.session, payload)
